import logging
import copy
import numpy as np

from .hyperband import HyperbandScheduler, HyperbandBracketManager
from .mo_asha_promotion import MOPromotionRungSystem, EPS_NET, NSGA_II
from ..utils.default_arguments import Integer, Boolean, Categorical, \
    filter_by_key
from ..utils.mo_hbo_utils import prepare_sign_vector
from ..utils.mo_hbo_utils import retrieve_pareto_front

__all__ = ['MOASHAScheduler']

logger = logging.getLogger(__name__)

_ARGUMENT_KEYS = {
    'objectives',
}


class MOASHAScheduler(HyperbandScheduler):
    r"""Implements different variants of asynchronous multi-objective Hyperband
    by extending the standard HyperbandScheduler using a multi-objective
    version of ASHA. This version internally ranks candidates by first sorting
    them into different Pareto frontiers and then spaces them out evenly based
    on euclidean distance.

    Parameters
    ----------
    objectives : dict
        Dictionary with the names of objectives of interest. The corresponding
        values are allowed to be either "MAX" or "MIN" and indicate if an
        objective is to be maximized or minimized.

    Examples
    --------
    examples/mo_hpo/mo_asha.ipynb
    """

    _CONSTRAINTS = {
        'resume': Boolean(),
        'max_t': Integer(1, None),
        'grace_period': Integer(1, None),
        'reduction_factor': Integer(2, None),
        'brackets': Integer(1, None),
        'type': Categorical((EPS_NET, NSGA_II)),
        'searcher_data': Categorical(
            ('rungs', 'all', 'rungs_and_last')),
        'do_snapshots': Boolean(),
        'rung_system_per_bracket': Boolean(),
        'random_seed': Integer(0, None)}

    def __init__(self, train_fn, **kwargs):
        self._objectives = kwargs["objectives"]
        assert kwargs["type"] in self._CONSTRAINTS["type"].choices, \
            "Only eps_net and nsga_ii are supported"
        # NOTE: This is just a dummy objective
        kwargs["reward_attr"] = list(self._objectives.keys())[0]
        super().__init__(train_fn, **filter_by_key(kwargs, _ARGUMENT_KEYS))

    def _init_bracket_manager(self, scheduler_type, max_t, rung_levels,
                              brackets, rung_system_per_bracket, random_seed):
        # Initialize bracket manager.
        terminator = MOHyperbandBracketManager(
            scheduler_type, self._time_attr, self._reward_attr, max_t,
            rung_levels, brackets, rung_system_per_bracket,
            random_seed, self._objectives)
        return terminator

    def get_pareto_front(self):
        """Retrieves the pareto efficient points discovered during the search
        process. Raises an error if called before search was conducted.

        Returns
        ----------
        front: list
            A list containing the Pareto efficient points among all points
            that were found during the search process.
        """
        assert len(self.finished_tasks) > 0, "Can only extract front after \
            jobs have been completed."
        pareto_front = retrieve_pareto_front(self.training_history,
                                             self._objectives)
        return pareto_front


class MOHyperbandBracketManager(HyperbandBracketManager):
    """MO-Hyperband Bracket Manager

    Arguments follow the parent class

    Args:
        objectives : dict
            Indicates objectives of interest for the rung system
    """
    def __init__(
            self, scheduler_type, time_attr, reward_attr, max_t, rung_levels,
            brackets, rung_system_per_bracket, random_seed, objectives):
        # Sign vector ensures that we deal with a pure minimization problem
        self.sign_vector = -prepare_sign_vector(objectives)
        self.objectives = objectives
        super().__init__(scheduler_type, time_attr, reward_attr, max_t,
                         rung_levels, brackets, rung_system_per_bracket,
                         random_seed)

    @staticmethod
    def _init_rung_systems(scheduler_type, rung_levels, promote_quantiles,
                           max_t, num_systems):
        # Initialize MOPromotionRungSystem rung system for each level.
        rung_systems = [
            MOPromotionRungSystem(rung_levels[s:], promote_quantiles[s:], max_t, scheduler_type)
            for s in range(num_systems)
        ]
        return rung_systems

    def _prepare_objective_vector(self, result):
        values = np.array([result[k] for k in self.objectives])
        values = values * self.sign_vector
        return values

    def on_task_report(self, task, result):
        """Passes a modified result dictionary to the parent class which causes
        the bracket manager to pass a numpy array containing multiple objective
        values to the rung system.

        :param task: Only task.task_id is used
        :param result: Current reported results from task
        :return: See parent class
        """
        result_copy = copy.copy(result)
        result_copy["_REWARD_VECTOR"] = \
            self._prepare_objective_vector(result_copy)
        return super().on_task_report(task, result_copy)

    def _report_to_rung_sys(self, rung_sys, task, result, skip_rungs):
        # Pass result to rung system to decide if task may continue.
        rung_info = rung_sys.on_task_report(
                task, result[self._time_attr], result["_REWARD_VECTOR"],
                skip_rungs=skip_rungs)
        return rung_info



import logging
import multiprocessing as mp
import os
import sys
from functools import partial

import dill

from autogluon.core.scheduler.managers import TaskManagers
from autogluon.core.scheduler.reporter import Communicator, LocalStatusReporter
from autogluon.core.utils import CustomProcess, AutoGluonEarlyStop
from autogluon.core.utils.files import make_temp_directory
from autogluon.common.utils.multiprocessing_utils import is_fork_enabled

logger = logging.getLogger(__name__)

__all__ = ['DistributedJobRunner']

SYS_ERR_OUT_FILE = 'sys_err.out'
SYS_STD_OUT_FILE = 'sys_std.out'


class DistributedJobRunner(object):

    @classmethod
    def start_distributed_job(cls, task, manager: TaskManagers):
        """Async Execute the job in remote and release the resources
        """
        logger.debug('\nScheduling {}'.format(task))
        job = task.resources.node.submit(cls._run_dist_job, task.task_id, task.fn, task.args, task.resources.gpu_ids)

        def _release_resource_callback(fut):
            logger.debug('Start Releasing Resource')
            manager.release_resources(task.resources)

        job.add_done_callback(_release_resource_callback)
        return job

    @classmethod
    def _worker(cls, tempdir, task_id, pickled_fn, pickled_args, return_list, gpu_ids, args):
        """Worker function in the client
        """

        with open(os.path.join(tempdir, f'{task_id}.out'), 'w') as std_out:
            with open(os.path.join(tempdir, f'{task_id}.err'), 'w') as err_out:

                # redirect stdout/strerr into a file so the main process can read it after the job is completed
                if not is_fork_enabled():
                    sys.stdout = std_out
                    sys.stderr = err_out

                # Only fork mode allows passing non-picklable objects
                fn = pickled_fn if is_fork_enabled() else dill.loads(pickled_fn)
                args = {**pickled_args, **args} if is_fork_enabled() else {**dill.loads(pickled_args), **args}

                DistributedJobRunner.set_cuda_environment(gpu_ids)

                # running
                try:
                    ret = fn(**args)
                except AutoGluonEarlyStop:
                    ret = None
                return_list.append(ret)

                sys.stdout.flush()
                sys.stderr.flush()

    @classmethod
    def set_cuda_environment(cls, gpu_ids):
        if len(gpu_ids) > 0:
            # handle GPU devices
            os.environ['CUDA_VISIBLE_DEVICES'] = ",".join(map(str, gpu_ids))
            os.environ['MXNET_CUDNN_AUTOTUNE_DEFAULT'] = "0"
        else:
            os.environ['CUDA_VISIBLE_DEVICES'] = "-1"


    @classmethod
    def _run_dist_job(cls, task_id, fn, args, gpu_ids):
        """Remote function Executing the task
        """
        if '_default_config' in args['args']:
            args['args'].pop('_default_config')

        if 'reporter' in args:
            local_reporter = LocalStatusReporter()
            dist_reporter = args['reporter']
            args['reporter'] = local_reporter

        manager = mp.Manager()
        return_list = manager.list()

        try:
            # Starting local process
            # Note: we have to use dill here because every argument passed to a child process over spawn or forkserver
            # has to be pickled. fork mode does not require this because memory sharing, but it is unusable for CUDA
            # applications (CUDA does not support fork) and multithreading issues (hanged threads).
            # Usage of decorators makes standard pickling unusable (https://docs.python.org/3/library/pickle.html#what-can-be-pickled-and-unpickled)
            # Dill enables sending of decorated classes. Please note if some classes are used in the training function,
            # those classes are best be defined inside the function - this way those can be constructed 'on-the-other-side'
            # after deserialization.
            pickled_fn = fn if is_fork_enabled() else dill.dumps(fn)

            # Reporter has to be separated since it's used for cross-process communication and has to be passed as-is
            args_ = {k: v for (k, v) in args.items() if k not in ['reporter']}
            pickled_args = args_ if is_fork_enabled() else dill.dumps(args_)

            cross_process_args = {k: v for (k, v) in args.items() if k not in ['fn', 'args']}

            with make_temp_directory() as tempdir:
                p = CustomProcess(
                    target=partial(cls._worker, tempdir, task_id, pickled_fn, pickled_args),
                    args=(return_list, gpu_ids, cross_process_args)
                )
                p.start()
                if 'reporter' in args:
                    cp = Communicator.Create(p, local_reporter, dist_reporter)
                p.join()
                # Get processes outputs
                if not is_fork_enabled():
                    cls.__print(tempdir, task_id, 'out')
                    cls.__print(tempdir, task_id, 'err')
        except Exception as e:
            logger.error('Exception in worker process: {}'.format(e))
        ret = return_list[0] if len(return_list) > 0 else None
        return ret

    @classmethod
    def __print(cls, tempdir, task_id, out):
        with open(os.path.join(tempdir, f'{task_id}.{out}')) as f:
            out = f.read()
            file = sys.stderr if out == 'err' else sys.stdout
            if out:
                print(f'(task:{task_id})\t{out}', file=file, end='')



from typing import Dict, List, Optional, Union
import numpy as np
import logging
from dataclasses import dataclass

from .model_base import BaseSurrogateModel
from ..datatypes.common import FantasizedPendingEvaluation, Candidate
from ..datatypes.hp_ranges_cs import HyperparameterRanges_CS
from ..datatypes.tuning_job_state import TuningJobState
from ..gpautograd.gp_regression import GaussianProcessRegression
from ..gpautograd.gpr_mcmc import GPRegressionMCMC
from ..gpautograd.posterior_state import GaussProcPosteriorState
from ..utils.debug_log import DebugLogPrinter
from ..utils.simple_profiler import SimpleProfiler
from ..tuning_algorithms.base_classes import DEFAULT_METRIC

logger = logging.getLogger(__name__)

GPModel = Union[GaussianProcessRegression, GPRegressionMCMC]

@dataclass
class InternalCandidateEvaluations:
    X: np.ndarray
    y: np.ndarray
    mean: float
    std: float


# Note: If state.pending_evaluations is not empty, it must contain entries
# of type FantasizedPendingEvaluation, which contain the fantasy samples. This
# is the case only for internal states, the member GPMXNetModel.state has
# PendingEvaluation entries without the fantasy samples.
def get_internal_candidate_evaluations(
        state: TuningJobState, active_metric: str, normalize_targets: bool,
        num_fantasize_samples: int) -> InternalCandidateEvaluations:
    candidates_ndarray = []
    evaluation_values = []
    for candidate_evaluation in state.candidate_evaluations:
        candidates_ndarray.append(
            state.hp_ranges.to_ndarray(candidate_evaluation.candidate))
        evaluation_values.append(candidate_evaluation.metrics[active_metric])
    X = np.vstack(candidates_ndarray)
    # Normalize
    # Note: The fantasy values in state.pending_evaluations are sampled
    # from the model fit to normalized targets, so they are already
    # normalized
    y = np.vstack(evaluation_values).reshape((-1, 1))
    mean = 0.0
    std = 1.0
    if normalize_targets:
        std = max(np.std(y).item(), 1e-15)
        mean = np.mean(y).item()
        y = (y - mean) / std
    if state.pending_evaluations:
        # In this case, y becomes a matrix, where the observed values are
        # broadcasted
        fanta_lst = []
        cand_lst = []
        for pending_eval in state.pending_evaluations:
            assert isinstance(pending_eval, FantasizedPendingEvaluation), \
                "state.pending_evaluations has to contain FantasizedPendingEvaluation"
            fantasies = pending_eval.fantasies[active_metric]
            assert fantasies.size == num_fantasize_samples, \
                "All state.pending_evaluations entries must have length {}".format(
                    num_fantasize_samples)
            fanta_lst.append(fantasies.reshape((1, -1)))
            cand_lst.append(state.hp_ranges.to_ndarray(pending_eval.candidate))
        y = np.vstack([y * np.ones((1, num_fantasize_samples))] + fanta_lst)
        X = np.vstack([X] + cand_lst)
    return InternalCandidateEvaluations(X, y, mean, std)


class GaussProcSurrogateModel(BaseSurrogateModel):
    def __init__(
            self, state: TuningJobState, active_metric: str, random_seed: int,
            gpmodel: GPModel, fit_parameters: bool, num_fantasy_samples: int,
            normalize_targets: bool = True,
            profiler: SimpleProfiler = None,
            debug_log: Optional[DebugLogPrinter] = None,
            debug_fantasy_values = None):
        """
        Given a TuningJobState state, the corresponding posterior state is
        computed here, based on which predictions are supported.
        Note: state is immutable. It must contain labeled examples.

        Parameters of the GP model in gpmodel are optimized iff fit_parameters
        is true. This requires state to contain labeled examples.

        We support pending evaluations via fantasizing. Note that state does
        not contain the fantasy values, but just the pending configs. Fantasy
        values are sampled here.

        :param state: TuningJobSubState
        :param active_metric: name of the metric to optimize.
        :param random_seed: Used only if GP model is created here
        :param gpmodel: GaussianProcessRegression model or GPRegressionMCMC model
        :param fit_parameters: Optimize parameters of gpmodel? Otherwise, these
            parameters are not changed
        :param num_fantasy_samples: See above
        :param normalize_targets: Normalize target values in
            state.candidate_evaluations?

        """
        super().__init__(state, active_metric, random_seed, debug_log)
        assert num_fantasy_samples > 0
        self._gpmodel = gpmodel
        self.num_fantasy_samples = num_fantasy_samples
        self.normalize_targets = normalize_targets
        self.active_metric = active_metric
        # Compute posterior (including fitting (optional) and dealing with
        # pending evaluations)
        # If state.pending_evaluations is not empty, fantasy samples are drawn
        # here, but they are not maintained in the state (which is immutable),
        # and not in the posterior state either.
        # Instead, fantasy samples can be accessed via self.fantasy_samples.
        self.fantasy_samples = None
        # DEBUG: Allows for comparison testing. If given, these fantasy values are
        # used in _draw_fantasy_values (instead of sampling them)
        self._debug_fantasy_values = debug_fantasy_values  # DEBUG
        self._compute_posterior(fit_parameters, profiler)

    def predict(self, inputs: np.ndarray) -> List[Dict[str, np.ndarray]]:
        """
        Note: Different to GPyOpt, means and stddevs are de-normalized here.
        """
        predictions_list_denormalized = []
        for posterior_mean, posterior_variance in self._gpmodel.predict(
                inputs):
            assert posterior_mean.shape[0] == inputs.shape[0], \
                (posterior_mean.shape, inputs.shape)
            assert posterior_variance.shape == (inputs.shape[0],), \
                (posterior_variance.shape, inputs.shape)
            if self.state.pending_evaluations:
                # If there are pending candidates with fantasy values,
                # posterior_mean must be a matrix
                assert posterior_mean.ndim == 2 and \
                    posterior_mean.shape[1] == self.num_fantasy_samples, \
                    (posterior_mean.shape, self.num_fantasy_samples)
            mean_denorm = posterior_mean * self.std + self.mean
            std_denorm = np.sqrt(posterior_variance) * self.std
            predictions_list_denormalized.append(
                {'mean': mean_denorm, 'std': std_denorm})
        return predictions_list_denormalized

    def backward_gradient(
            self, input: np.ndarray,
            head_gradients: List[Dict[str, np.ndarray]]) -> List[np.ndarray]:
        poster_states = self.posterior_states()
        assert poster_states is not None, \
            "Cannot run backward_gradient without a posterior state"
        assert len(poster_states) == len(head_gradients), \
            "len(posterior_states) = {} != {} = len(head_gradients)".format(
                len(poster_states), len(head_gradients))
        return [
            poster_state.backward_gradient(
                input, head_gradient, self.mean, self.std)
            for poster_state, head_gradient in zip(poster_states, head_gradients)
        ]

    @property
    def gpmodel(self) -> GPModel:
        return self._gpmodel

    def does_mcmc(self):
        return isinstance(self._gpmodel, GPRegressionMCMC)

    def posterior_states(self) -> Optional[List[GaussProcPosteriorState]]:
        return self._gpmodel.states

    def get_params(self):
        """
        Note: Once MCMC is supported, this method will have to be refactored.
        Note: If self.state still has no labeled data, the parameters returned
        are the initial ones, where an update would start from.

        :return: Hyperparameter dictionary
        """
        if not self.does_mcmc():
            return self._gpmodel.get_params()
        else:
            return dict()

    def set_params(self, param_dict):
        self._gpmodel.set_params(param_dict)

    def _compute_posterior(
            self, fit_parameters: bool, profiler: SimpleProfiler):
        """
        Completes __init__, by computing the posterior. If fit_parameters, this
        includes optimizing the surrogate model parameters.

        If self.state.pending_evaluations is not empty, we proceed as follows:
        - Compute posterior for state without pending evals
        - Draw fantasy values for pending evals
        - Recompute posterior (without fitting)

        """
        if self._debug_log is not None:
            self._debug_log.set_state(self.state)
        # Compute posterior for state without pending evals
        no_pending_state = self.state
        if self.state.pending_evaluations:
            no_pending_state = TuningJobState(
                hp_ranges=self.state.hp_ranges,
                candidate_evaluations=self.state.candidate_evaluations,
                failed_candidates=self.state.failed_candidates,
                pending_evaluations=[])
        self._posterior_for_state(no_pending_state, fit_parameters, profiler)
        if self.state.pending_evaluations:
            # Sample fantasy values for pending evals
            pending_configs = [
                x.candidate for x in self.state.pending_evaluations]
            new_pending = self._draw_fantasy_values(pending_configs)
            # Compute posterior for state with pending evals
            # Note: profiler is not passed here, this would overwrite the
            # results from the first call
            with_pending_state = TuningJobState(
                hp_ranges=self.state.hp_ranges,
                candidate_evaluations=self.state.candidate_evaluations,
                failed_candidates=self.state.failed_candidates,
                pending_evaluations=new_pending)
            self._posterior_for_state(
                with_pending_state, fit_parameters=False, profiler=None)
            # Note: At this point, the fantasy values are dropped, they are not
            # needed anymore. They've just been sampled for the posterior
            # computation. We still maintain them in self.fantasy_samples,
            # which is mainly used for testing
            self.fantasy_samples = new_pending

    def _posterior_for_state(
            self, state: TuningJobState, fit_parameters: bool,
            profiler: Optional[SimpleProfiler]):
        """
        Computes posterior for state.
        If fit_parameters and state.pending_evaluations is empty, we first
        optimize the model parameters.
        If state.pending_evaluations are given, these must be
        FantasizedPendingEvaluations, i.e. the fantasy values must have been
        sampled.

        """
        assert state.candidate_evaluations, \
            "Cannot compute posterior: state has no labeled datapoints"
        internal_candidate_evaluations = get_internal_candidate_evaluations(
            state, self.active_metric, self.normalize_targets,
            self.num_fantasy_samples)
        X_all = internal_candidate_evaluations.X
        Y_all = internal_candidate_evaluations.y
        assert X_all.shape[0] == Y_all.shape[0]
        self.mean = internal_candidate_evaluations.mean
        self.std = internal_candidate_evaluations.std

        fit_parameters = fit_parameters and (not state.pending_evaluations)
        if not fit_parameters:
            logger.log(15, "Recomputing GP state")
            self._gpmodel.recompute_states(X_all, Y_all, profiler=profiler)
        else:
            logger.log(15, f"Fitting GP model for {self.active_metric}")
            self._gpmodel.fit(X_all, Y_all, profiler=profiler)
        if self._debug_log is not None:
            self._debug_log.set_gp_params(self.get_params())
            if not state.pending_evaluations:
                deb_msg = "[GaussProcSurrogateModel._posterior_for_state]\n"
                deb_msg += ("- self.mean = {}\n".format(self.mean))
                deb_msg += ("- self.std = {}".format(self.std))
                logger.info(deb_msg)
                self._debug_log.set_targets(internal_candidate_evaluations.y)
            else:
                num_pending = len(state.pending_evaluations)
                fantasies = internal_candidate_evaluations.y[-num_pending:, :]
                self._debug_log.set_fantasies(fantasies)

    def _draw_fantasy_values(self, candidates: List[Candidate]) \
            -> List[FantasizedPendingEvaluation]:
        """
        Note: The fantasy values need not be de-normalized, because they are
        only used internally here (e.g., get_internal_candidate_evaluations).

        Note: A complication is that if the sampling methods of _gpmodel
        are called when there are no pending candidates (with fantasies) yet,
        they do return a single sample (instead of num_fantasy_samples). This
        is because GaussianProcessRegression knows about num_fantasy_samples
        only due to the form of the posterior state (bad design!).
        In this case, we draw num_fantasy_samples i.i.d.

        """
        if candidates:
            if self._debug_fantasy_values is not None:
                # DEBUG: Use provided fantasy values
                assert len(self._debug_fantasy_values) == len(candidates)
                logger.info("DEBUG: Use given fantasy values, rather than resampling them")
                return self._debug_fantasy_values

            logger.debug("Fantasizing target values for candidates:\n{}"
                         .format(candidates))
            X_new = self.state.hp_ranges.to_ndarray_matrix(candidates)
            # Special case (see header comment): If the current posterior state
            # does not contain pending candidates (no fantasies), we sample
            # num_fantasy_samples times i.i.d.
            num_samples = 1 if self._gpmodel.multiple_targets() \
                else self.num_fantasy_samples
            # We need joint sampling for >1 new candidates
            num_candidates = len(candidates)
            sample_func = self._gpmodel.sample_joint if num_candidates > 1 else \
                self._gpmodel.sample_marginals
            Y_new = sample_func(X_new, num_samples=num_samples).reshape(
                (num_candidates, -1))
            return [
                FantasizedPendingEvaluation(
                    candidate, {self.active_metric: y_new.reshape((1, -1))})
                for candidate, y_new in zip(candidates, Y_new)
            ]
        else:
            return []

    def current_best_filter_candidates(self, candidates):
        hp_ranges = self.state.hp_ranges
        if isinstance(hp_ranges, HyperparameterRanges_CS):
            candidates = hp_ranges.filter_for_last_pos_value(candidates)
            assert candidates, \
                "state.hp_ranges does not contain any candidates " + \
                "(labeled or pending) with resource attribute " + \
                "'{}' = {}".format(
                    hp_ranges.name_last_pos, hp_ranges.value_for_last_pos)
        return candidates


# Convenience type allowing for multi-output HPO. This is used for methods that work both in the standard case
# of a single output model and in the multi-output case (e.g., see GPModelPendingCandidateStateTransformer)
GaussProcSurrogateOutputModel = Union[GaussProcSurrogateModel, Dict[str, GaussProcSurrogateModel]]


@dataclass
class GPModelArgs:
    num_fantasy_samples: int
    random_seed: int
    active_metric: str = DEFAULT_METRIC
    normalize_targets: bool = True


GPModelArgsOutput = Union[GPModelArgs, Dict[str, GPModelArgs]]
GPOutputModel = Union[GPModel, Dict[str, GPModel]]


from typing import Callable, Dict, Type, Optional, Any
import copy
import logging

from .gp_fifo_searcher import GPFIFOSearcher, decode_state, \
    MapReward, GET_CONFIG_RANDOM_RETRIES, create_initial_candidates_scorer
from ..datatypes.common import CandidateEvaluation, Candidate, \
    candidate_for_print
from ..datatypes.hp_ranges import HyperparameterRanges
from ..datatypes.tuning_job_state import TuningJobState
from ..models.gp_model import GPModel, GPModelArgs
from ..models.gpmodel_skipopt import SkipOptimizationPredicate
from ..models.gpmodel_transformers import GPModelPendingCandidateStateTransformer
from ..tuning_algorithms.base_classes import DEFAULT_METRIC, DEFAULT_CONSTRAINT_METRIC, \
    LocalOptimizer, AcquisitionFunction
from ..tuning_algorithms.bo_algorithm import BayesianOptimizationAlgorithm
from ..tuning_algorithms.common import compute_blacklisted_candidates
from ..tuning_algorithms.defaults import DEFAULT_LOCAL_OPTIMIZER_CLASS, DEFAULT_NUM_INITIAL_CANDIDATES, \
    DEFAULT_NUM_INITIAL_RANDOM_EVALUATIONS
from ..utils.debug_log import DebugLogPrinter
from ..utils.duplicate_detector import DuplicateDetectorIdentical
from ..utils.simple_profiler import SimpleProfiler

logger = logging.getLogger(__name__)


class ConstrainedGPFIFOSearcher(GPFIFOSearcher):
    """
    Supports constrained GP-based hyperparameter optimization (to be used with a FIFO scheduler).

    """
    def __init__(self, hp_ranges: HyperparameterRanges, random_seed: int,
                 output_gpmodels: Dict[str, GPModel],
                 output_models_args: Dict[str, GPModelArgs],
                 map_reward: MapReward,
                 acquisition_class: Type[AcquisitionFunction],
                 init_state: TuningJobState = None,
                 local_minimizer_class: Type[LocalOptimizer] = DEFAULT_LOCAL_OPTIMIZER_CLASS,
                 skip_optimization: SkipOptimizationPredicate = None,
                 num_initial_candidates: int = DEFAULT_NUM_INITIAL_CANDIDATES,
                 num_initial_random_choices: int = DEFAULT_NUM_INITIAL_RANDOM_EVALUATIONS,
                 initial_scoring: Optional[str] = None,
                 first_is_default: bool = True,
                 debug_log: Optional[DebugLogPrinter] = None,
                 cost_metric_name: Optional[str] = None,
                 profiler: Optional[SimpleProfiler] = None,
                 getconfig_callback: Optional[Callable[[dict], Any]] = None):
        """
        Note that the SurrogateModel is created on demand (by the state
        transformer) in get_config, along with components needed for the BO
        algorithm.

        The searcher is supposed to maximize reward, while internally, the
        criterion is minimized. map_reward maps reward to internal criterion, it
        must be strictly decreasing.

        :param hp_ranges: Configuration space without resource attribute
        :param random_seed:
        :param output_gpmodels: Dict with the GP regression model for the active metric and constraint metric
        :param output_models_args: Dict with arguments for GPMXNet model creation for the active and constraint metric
        :param map_reward: Function mapping reward to criterion to be minimized
        :param acquisition_class: Type for acquisition function
        :param init_state: TuningJobState to start from (default is empty)
        :param local_minimizer_class: Type for local minimizer
        :param skip_optimization: Predicate, see
            GPMXNetPendingCandidateStateTransformer
        :param num_initial_candidates: See BayesianOptimizationAlgorithm
        :param num_initial_random_choices: Configs are sampled at random until
            this many candidates received label feedback
        :param initial_scoring: Scoring function to rank initial candidates.
            Default: thompson_indep (independent Thompson sampling)
        :param first_is_default: If true, the first config to be evaluated is
            the default of the search space. Otherwise, the first is sampled
            at random
        :param debug_log: DebugLogPrinter for debug logging (optional)
        :param cost_metric_name: If an `elapsed_time` record is passed to
            `update`, the value is entered in the state using this key
        :param profiler: If given, HPO computations are profiled
        :param getconfig_callback: If given, this callback function is called
            at the end of each get_config. It receives the chosen config
            (as dict) as single argument. One use case for this callback is
            to store profiler records.
        """
        super().__init__(hp_ranges, random_seed, output_gpmodels[DEFAULT_METRIC],
                         output_models_args[DEFAULT_METRIC], map_reward, acquisition_class, init_state,
                         local_minimizer_class, skip_optimization, num_initial_candidates, num_initial_random_choices,
                         initial_scoring, first_is_default, debug_log, cost_metric_name, profiler, getconfig_callback)

        self.skip_optimization = skip_optimization
        self.state_transformer = GPModelPendingCandidateStateTransformer(
            gpmodel=output_gpmodels,
            init_state=self._init_state,
            model_args=output_models_args,
            skip_optimization=skip_optimization,
            profiler=profiler,
            debug_log=debug_log)

        if debug_log is not None:
            deb_msg = "[ConstrainedGPFIFOSearcher.__init__]\n"
            deb_msg += ("- acquisition_class = {}\n".format(acquisition_class))
            deb_msg += ("- local_minimizer_class = {}\n".format(local_minimizer_class))
            deb_msg += ("- num_initial_candidates = {}\n".format(num_initial_candidates))
            deb_msg += ("- num_initial_random_choices = {}\n".format(num_initial_random_choices))
            deb_msg += ("- initial_scoring = {}\n".format(self.initial_scoring))
            deb_msg += ("- first_is_default = {}".format(first_is_default))
            logger.info(deb_msg)

    def update(self, config: Candidate, reward: float, constraint: float, **kwargs):
        """
        Registers new datapoint at config, with reward reward.
        Note that in general, config should previously have been registered as
        pending (register_pending). If so, it is switched from pending
        to labeled. If not, it is considered directly labeled.
        """
        crit_val = self.map_reward(reward)
        constr_val = constraint
        metrics = {DEFAULT_METRIC: crit_val,
                   DEFAULT_CONSTRAINT_METRIC: constr_val}
        if 'elapsed_time' in kwargs:
            metrics[self.cost_metric_name] = kwargs['elapsed_time']
        self.state_transformer.label_candidate(CandidateEvaluation(
            candidate=copy.deepcopy(config), metrics=metrics))
        if self.debug_log is not None:
            config_id = self.debug_log.config_id(config)
            msg = "Update for config_id {}: reward = {}, crit_val = {}, constr_val = {}".format(
                config_id, reward, crit_val, constr_val)
            logger.info(msg)

    def get_config(self) -> Candidate:
        """
        Runs Bayesian optimization in order to suggest the next config to evaluate.

        :return: Next config to evaluate at
        """
        state = self.state_transformer.state
        if self.do_profile:
            # Start new profiler block
            meta = {
                'fit_hyperparams': not self.state_transformer.skip_optimization(
                    state),
                'num_observed': len(state.candidate_evaluations),
                'num_pending': len(state.pending_evaluations)
            }
            self.profiler.begin_block(meta)
            self.profiler.start('all')
        blacklisted_candidates = compute_blacklisted_candidates(state)
        pick_random = (len(blacklisted_candidates) < self.num_initial_random_choices) or \
            (not state.candidate_evaluations)
        if self.debug_log is not None:
            self.debug_log.start_get_config('random' if pick_random else 'BO')
        if pick_random:
            config = None
            if self.first_is_default and (not blacklisted_candidates):
                # Use default configuration if there is one specified
                default_config = self.hp_ranges.config_space.get_default_configuration()
                if default_config and len(default_config.get_dictionary()) > 0:
                    config = default_config
                    if self.debug_log is not None:
                        logger.info("Start with default config:\n{}".format(
                            candidate_for_print(config)))
            if config is None:
                if self.do_profile:
                    self.profiler.start('random')
                for _ in range(GET_CONFIG_RANDOM_RETRIES):
                    _config = self.hp_ranges.random_candidate(self.random_state)
                    if _config not in blacklisted_candidates:
                        config = _config
                        break
                if self.do_profile:
                    self.profiler.stop('random')
                if config is None:
                    raise AssertionError(
                        "Failed to sample a configuration not already chosen "
                        "before. Maybe there are no free configurations left? "
                        "The blacklist size is {}".format(len(blacklisted_candidates)))
        else:
            # Obtain current SurrogateModel from state transformer. Based on
            # this, the BO algorithm components can be constructed
            if self.do_profile:
                self.profiler.push_prefix('getconfig')
                self.profiler.start('all')
                self.profiler.start('gpmodel')
            # Create the active and constraint models
            output_models = self.state_transformer.model()  # Asking for the model triggers the posterior computation
            if self.do_profile:
                self.profiler.stop('total_update')
            # Create BO algorithm
            initial_candidates_scorer = create_initial_candidates_scorer(initial_scoring=self.initial_scoring,
                                                                         model=output_models,
                                                                         acquisition_class=self.acquisition_class,
                                                                         random_state=self.random_state,
                                                                         active_output=DEFAULT_METRIC)
            local_optimizer = self.local_minimizer_class(state=state,
                                                         model=output_models,
                                                         acquisition_function_class=self.acquisition_class,
                                                         active_metric=DEFAULT_METRIC)
            bo_algorithm = BayesianOptimizationAlgorithm(
                initial_candidates_generator=self.random_generator,
                initial_candidates_scorer=initial_candidates_scorer,
                num_initial_candidates=self.num_initial_candidates,
                local_optimizer=local_optimizer,
                pending_candidate_state_transformer=None,
                blacklisted_candidates=blacklisted_candidates,
                num_requested_candidates=1,
                greedy_batch_selection=False,
                duplicate_detector=DuplicateDetectorIdentical(),
                profiler=self.profiler,
                sample_unique_candidates=False,
                debug_log=self.debug_log)
            # Next candidate decision
            _config = bo_algorithm.next_candidates()
            if len(_config) == 0:
                raise AssertionError(
                    "Failed to find a configuration not already chosen "
                    "before. Maybe there are no free configurations left? "
                    "The blacklist size is {}".format(len(blacklisted_candidates)))
            config = _config[0]
            if self.do_profile:
                self.profiler.stop('all')
                self.profiler.pop_prefix()  # getconfig

        if self.debug_log is not None:
            self.debug_log.set_final_config(config)
            # All get_config debug log info is only written here
            self.debug_log.write_block()
        if self.do_profile:
            self.profiler.stop('all')
            self.profiler.clear()
        if self.getconfig_callback is not None:
            self.getconfig_callback(config.get_dictionary())

        return config

    def clone_from_state(self, state):
        # Create clone with mutable state taken from 'state'
        output_models_args = self.state_transformer._model_args
        output_gpmodels = self.state_transformer._gpmodel
        init_state = decode_state(state['state'], self.hp_ranges)
        skip_optimization = state['skip_optimization']
        new_searcher = ConstrainedGPFIFOSearcher(
            hp_ranges=self.hp_ranges,
            random_seed=self.random_seed,
            output_gpmodels=output_gpmodels,
            output_models_args=output_gpmodels,
            map_reward=self.map_reward,
            acquisition_class=self.acquisition_class,
            init_state=init_state,
            local_minimizer_class=self.local_minimizer_class,
            skip_optimization=skip_optimization,
            num_initial_candidates=self.num_initial_candidates,
            num_initial_random_choices=self.num_initial_random_choices,
            initial_scoring=self.initial_scoring,
            profiler=self.profiler,
            first_is_default=self.first_is_default,
            debug_log=self.debug_log)
        new_searcher.state_transformer.set_params(state['model_params'])
        new_searcher.random_state = state['random_state']
        new_searcher.random_generator.random_state = \
            state['random_state']
        if self.debug_log and 'debug_log' in state:
            new_searcher.debug_log.set_mutable_state(state['debug_log'])
        # Invalidate self (must not be used afterwards)
        self.state_transformer = None
        return new_searcher
